import unittest

from robot.parsing.model.statements import *
from robot.parsing import Token
from robot.utils.asserts import assert_equal, assert_true
from robot.utils import type_name


def assert_created_statement(tokens, base_class, **params):
    statement = base_class.from_params(**params)
    assert_statements(
        statement,
        base_class(tokens)
    )
    assert_statements(
        statement,
        base_class.from_tokens(tokens)
    )
    assert_statements(
        statement,
        Statement.from_tokens(tokens)
    )
    if len(set(id(t) for t in statement.tokens)) != len(tokens):
        lines = '\n'.join(f'{i:18}{t}' for i, t in
                          [('ID', 'TOKEN')] +
                          [(str(id(t)), repr(t)) for t in statement.tokens])
        raise AssertionError(f'Tokens should not be reused!\n\n{lines}')
    return statement


def compare_statements(first, second):
    return (isinstance(first, type(second))
            and first.tokens == second.tokens
            and first.errors == second.errors)


def assert_statements(st1, st2):
    assert_equal(len(st1), len(st2),
                 f'Statement lengths are not equal:\n'
                 f'{len(st1)} for {st1}\n'
                 f'{len(st2)} for {st2}')
    for t1, t2 in zip(st1, st2):
        assert_equal(t1, t2, formatter=repr)
    assert_true(compare_statements(st1, st2),
                f'Statements are not equal:\n'
                f'{st1} {type_name(st1)}\n'
                f'{st2} {type_name(st2)}')


class TestStatementFromTokens(unittest.TestCase):

    def test_keyword_call_with_assignment(self):
        tokens = [Token(Token.SEPARATOR, '  '),
                  Token(Token.ASSIGN, '${var}'),
                  Token(Token.SEPARATOR, '  '),
                  Token(Token.KEYWORD, 'Keyword'),
                  Token(Token.SEPARATOR, '  '),
                  Token(Token.ARGUMENT, 'arg'),
                  Token(Token.EOL)]
        assert_statements(Statement.from_tokens(tokens), KeywordCall(tokens))

    def test_inline_if_with_assignment(self):
        tokens = [Token(Token.SEPARATOR, '  '),
                  Token(Token.ASSIGN, '${var}'),
                  Token(Token.SEPARATOR, '  '),
                  Token(Token.INLINE_IF, 'IF'),
                  Token(Token.SEPARATOR, '  '),
                  Token(Token.ARGUMENT, 'True'),
                  Token(Token.EOL)]
        assert_statements(Statement.from_tokens(tokens), InlineIfHeader(tokens))

    def test_assign_only(self):
        tokens = [Token(Token.SEPARATOR, '  '),
                  Token(Token.ASSIGN, '${var}'),
                  Token(Token.EOL)]
        assert_statements(Statement.from_tokens(tokens), KeywordCall(tokens))


class TestCreateStatementsFromParams(unittest.TestCase):

    def test_Statement(self):
        self.assertRaises(NotImplementedError, Statement.from_params)

    def test_SectionHeader(self):
        headers = {
            Token.SETTING_HEADER: 'Settings',
            Token.VARIABLE_HEADER: 'Variables',
            Token.TESTCASE_HEADER: 'Test Cases',
            Token.TASK_HEADER: 'Tasks',
            Token.KEYWORD_HEADER: 'Keywords',
            Token.COMMENT_HEADER: 'Comments'
        }
        for token_type, name in headers.items():
            tokens = [
                Token(token_type, '*** %s ***' % name),
                Token(Token.EOL, '\n')
            ]
            assert_created_statement(
                tokens,
                SectionHeader,
                type=token_type,
            )
            assert_created_statement(
                tokens,
                SectionHeader,
                type=token_type,
                name=name
            )
            assert_created_statement(
                tokens,
                SectionHeader,
                type=token_type,
                name='*** %s ***' % name
            )

    def test_SuiteSetup(self):
        # Suite Setup    Setup Keyword    ${arg1}    ${arg2}
        tokens = [
            Token(Token.SUITE_SETUP, 'Suite Setup'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Setup Keyword'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg2}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            SuiteSetup,
            name='Setup Keyword',
            args=['${arg1}', '${arg2}']
        )

    def test_SuiteTeardown(self):
        # Suite Teardown    Teardown Keyword    ${arg1}    ${arg2}
        tokens = [
            Token(Token.SUITE_TEARDOWN, 'Suite Teardown'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Teardown Keyword'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg2}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            SuiteTeardown,
            name='Teardown Keyword',
            args=['${arg1}', '${arg2}']
        )

    def test_TestSetup(self):
        # Test Setup    Setup Keyword    ${arg1}    ${arg2}
        tokens = [
            Token(Token.TEST_SETUP, 'Test Setup'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Setup Keyword'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg2}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            TestSetup,
            name='Setup Keyword',
            args=['${arg1}', '${arg2}']
        )

    def test_TestTeardown(self):
        # Test Teardown    Teardown Keyword    ${arg1}    ${arg2}
        tokens = [
            Token(Token.TEST_TEARDOWN, 'Test Teardown'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Teardown Keyword'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg2}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            TestTeardown,
            name='Teardown Keyword',
            args=['${arg1}', '${arg2}']
        )

    def test_TestTemplate(self):
        # Test Template    Keyword Template
        tokens = [
            Token(Token.TEST_TEMPLATE, 'Test Template'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Keyword Template'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            TestTemplate,
            value='Keyword Template'
        )

    def test_TestTimeout(self):
        # Test Timeout    1 min
        tokens = [
            Token(Token.TEST_TIMEOUT, 'Test Timeout'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '1 min'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            TestTimeout,
            value='1 min'
        )

    def test_KeywordTags(self):
        # Keyword Tags    first    second
        tokens = [
            Token(Token.KEYWORD_TAGS, 'Keyword Tags'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'first'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'second'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            KeywordTags,
            values=['first', 'second']
        )

    def test_Variable(self):
        # ${variable_name}  {'a': 4, 'b': 'abc'}
        tokens = [
            Token(Token.VARIABLE, '${variable_name}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, "{'a': 4, 'b': 'abc'}"),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Variable,
            name='${variable_name}',
            value="{'a': 4, 'b': 'abc'}"
        )
        # ${var}    first    second    third
        # @{var}    first    second    third
        # &{var}    first    second    third
        for name in ['${var}', '@{var}', '&{var}']:
            tokens = [
                Token(Token.VARIABLE, name),
                Token(Token.SEPARATOR, '    '),
                Token(Token.ARGUMENT, 'first'),
                Token(Token.SEPARATOR, '    '),
                Token(Token.ARGUMENT, 'second'),
                Token(Token.SEPARATOR, '    '),
                Token(Token.ARGUMENT, 'third'),
                Token(Token.EOL, '\n')
            ]
            assert_created_statement(
                tokens,
                Variable,
                name=name,
                value=['first', 'second', 'third']
            )

    def test_TestCaseName(self):
        tokens = [Token(Token.TESTCASE_NAME, 'Example test case name'), Token(Token.EOL, '\n')]
        assert_created_statement(
            tokens,
            TestCaseName,
            name='Example test case name'
        )

    def test_KeywordName(self):
        tokens = [Token(Token.KEYWORD_NAME, 'Keyword Name With ${embedded} Var'), Token(Token.EOL, '\n')]
        assert_created_statement(
            tokens,
            KeywordName,
            name='Keyword Name With ${embedded} Var'
        )

    def test_Setup(self):
        # Test
        #     [Setup]    Setup Keyword    ${arg1}
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.SETUP, '[Setup]'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Setup Keyword'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Setup,
            name='Setup Keyword',
            args=['${arg1}']
        )

    def test_Teardown(self):
        # Test
        #     [Teardown]    Teardown Keyword    ${arg1}
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.TEARDOWN, '[Teardown]'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Teardown Keyword'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Teardown,
            name='Teardown Keyword',
            args=['${arg1}']
        )

    def test_LibraryImport(self):
        # Library    library_name.py
        tokens = [
            Token(Token.LIBRARY, 'Library'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'library_name.py'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            LibraryImport,
            name='library_name.py'
        )

        # Library    library_name.py    WITH NAME    anothername
        tokens = [
            Token(Token.LIBRARY, 'Library'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'library_name.py'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.WITH_NAME),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'anothername'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            LibraryImport,
            name='library_name.py',
            alias='anothername'
        )

    def test_ResourceImport(self):
        # Resource    path${/}to${/}resource.robot
        tokens = [
            Token(Token.RESOURCE, 'Resource'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'path${/}to${/}resource.robot'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ResourceImport,
            name='path${/}to${/}resource.robot'
        )

    def test_VariablesImport(self):
        # Variables    variables.py
        tokens = [
            Token(Token.VARIABLES, 'Variables'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'variables.py'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            VariablesImport,
            name='variables.py'
        )

        # Variables    variables.py    arg1    2
        tokens = [
            Token(Token.VARIABLES, 'Variables'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'variables.py'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'arg1'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '2'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            VariablesImport,
            name='variables.py',
            args=['arg1', '2']
        )

    def test_Documentation(self):
        # Documentation    Example documentation
        tokens = [
            Token(Token.DOCUMENTATION, 'Documentation'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'Example documentation'),
            Token(Token.EOL, '\n')
        ]
        doc = assert_created_statement(
            tokens,
            Documentation,
            value='Example documentation'
        )
        assert_equal(doc.value, 'Example documentation')

        # Documentation    First line.
        # ...              Second line aligned.
        # ...
        # ...              Second paragraph.
        tokens = [
            Token(Token.DOCUMENTATION, 'Documentation'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'First line.'),
            Token(Token.EOL),
            Token(Token.CONTINUATION),
            Token(Token.SEPARATOR, '              '),
            Token(Token.ARGUMENT, 'Second line aligned.'),
            Token(Token.EOL),
            Token(Token.CONTINUATION),
            Token(Token.ARGUMENT, ''),
            Token(Token.EOL),
            Token(Token.CONTINUATION),
            Token(Token.SEPARATOR, '              '),
            Token(Token.ARGUMENT, 'Second paragraph.'),
            Token(Token.EOL),
        ]
        doc = assert_created_statement(
            tokens,
            Documentation,
            value='First line.\nSecond line aligned.\n\nSecond paragraph.'
        )
        assert_equal(doc.value, 'First line.\nSecond line aligned.\n\nSecond paragraph.')

        # Test/Keyword
        #     [Documentation]      First line
        #     ...                  Second line aligned
        #     ...
        #     ...                  Second paragraph.
        tokens = [
            Token(Token.SEPARATOR, '  '),
            Token(Token.DOCUMENTATION, '[Documentation]'),
            Token(Token.SEPARATOR, '      '),
            Token(Token.ARGUMENT, 'First line.'),
            Token(Token.EOL),
            Token(Token.SEPARATOR, '  '),
            Token(Token.CONTINUATION),
            Token(Token.SEPARATOR, '                  '),
            Token(Token.ARGUMENT, 'Second line aligned.'),
            Token(Token.EOL),
            Token(Token.SEPARATOR, '  '),
            Token(Token.CONTINUATION),
            Token(Token.ARGUMENT, ''),
            Token(Token.EOL),
            Token(Token.SEPARATOR, '  '),
            Token(Token.CONTINUATION),
            Token(Token.SEPARATOR, '                  '),
            Token(Token.ARGUMENT, 'Second paragraph.'),
            Token(Token.EOL),
        ]
        doc = assert_created_statement(
            tokens,
            Documentation,
            value='First line.\nSecond line aligned.\n\nSecond paragraph.\n',
            indent='  ',
            separator='      ',
            settings_section=False
        )
        assert_equal(doc.value, 'First line.\nSecond line aligned.\n\nSecond paragraph.')

    def test_Metadata(self):
        tokens = [
            Token(Token.METADATA, 'Metadata'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Key'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'Value'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Metadata,
            name='Key',
            value='Value'
        )

        tokens = [
            Token(Token.METADATA, 'Metadata'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Key'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'First line'),
            Token(Token.EOL, '\n'),
            Token(Token.CONTINUATION),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'Second line'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Metadata,
            name='Key',
            value='First line\nSecond line'
        )

    def test_Tags(self):
        # Test/Keyword
        #     [Tags]    tag1    tag2
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.TAGS, '[Tags]'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'tag1'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'tag2'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Tags,
            values=['tag1', 'tag2']
        )

    def test_ForceTags(self):
        tokens = [
            Token(Token.FORCE_TAGS, 'Force Tags'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'some tag'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'another_tag'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ForceTags,
            values=['some tag', 'another_tag']
        )

    def test_DefaultTags(self):
        tokens = [
            Token(Token.DEFAULT_TAGS, 'Default Tags'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'some tag'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'another_tag'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            DefaultTags,
            values=['some tag', 'another_tag']
        )

    def test_Template(self):
        # Test
        #     [Template]  Keyword Name
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.TEMPLATE, '[Template]'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.NAME, 'Keyword Name'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Template,
            value='Keyword Name'
        )

    def test_Timeout(self):
        # Test
        #     [Timeout]  1 min
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.TIMEOUT, '[Timeout]'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '1 min'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Timeout,
            value='1 min'
        )

    def test_Arguments(self):
        # Keyword
        #     [Arguments]    ${arg1}    ${arg2}=4
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENTS, '[Arguments]'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg2}=4'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Arguments,
            args=['${arg1}', '${arg2}=4']
        )

    def test_ReturnSetting(self):
        # Keyword
        #     [Return]    ${arg1}    ${arg2}=4
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.RETURN, '[Return]'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg2}=4'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ReturnSetting,
            args=['${arg1}', '${arg2}=4']
        )

    def test_KeywordCall(self):
        # Test
        #     ${return1}    ${return2}    Keyword Call    ${arg1}    ${arg2}
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.ASSIGN, '${return1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ASSIGN, '${return2}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.KEYWORD, 'Keyword Call'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg2}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            KeywordCall,
            name='Keyword Call',
            assign=['${return1}', '${return2}'],
            args=['${arg1}', '${arg2}']
        )

    def test_TemplateArguments(self):
        # Test
        #     [Template]   Templated Keyword
        #     ${arg1}    2
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${arg1}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '2'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            TemplateArguments,
            args=['${arg1}', '2']
        )

    def test_ForHeader(self):
        # Keyword
        #     FOR  ${value1}  ${value2}  IN ZIP  ${list1}  ${list2}
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.FOR),
            Token(Token.SEPARATOR, '  '),
            Token(Token.VARIABLE, '${value1}'),
            Token(Token.SEPARATOR, '  '),
            Token(Token.VARIABLE, '${value2}'),
            Token(Token.SEPARATOR, '  '),
            Token(Token.FOR_SEPARATOR, 'IN ZIP'),
            Token(Token.SEPARATOR, '  '),
            Token(Token.ARGUMENT, '${list1}'),
            Token(Token.SEPARATOR, '  '),
            Token(Token.ARGUMENT, '${list2}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ForHeader,
            flavor='IN ZIP',
            variables=['${value1}', '${value2}'],
            values=['${list1}', '${list2}'],
            separator='  '
        )

    def test_IfHeader(self):
        # Test/Keyword
        #     IF    ${var} not in [@{list}]
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.IF),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${var} not in [@{list}]'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            IfHeader,
            condition='${var} not in [@{list}]'
        )

    def test_InlineIfHeader(self):
        # Test/Keyword
        #     IF    $x > 0
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.INLINE_IF),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '$x > 0')
        ]
        assert_created_statement(
            tokens,
            InlineIfHeader,
            condition='$x > 0'
        )

    def test_InlineIfHeader_with_assign(self):
        # Test/Keyword
        #     ${y} =    IF    $x > 0
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.ASSIGN, '${y}'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.INLINE_IF),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '$x > 0')
        ]
        assert_created_statement(
            tokens,
            InlineIfHeader,
            condition='$x > 0',
            assign=['${y}']
        )

    def test_ElseIfHeader(self):
        # Test/Keyword
        #     ELSE IF    ${var} not in [@{list}]
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.ELSE_IF),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '${var} not in [@{list}]'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ElseIfHeader,
            condition='${var} not in [@{list}]'
        )

    def test_ElseHeader(self):
        # Test/Keyword
        #     ELSE
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.ELSE),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ElseHeader
        )

    def test_TryHeader(self):
        # TRY
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.TRY),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            TryHeader
        )

    def test_ExceptHeader(self):
        # EXCEPT
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.EXCEPT),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ExceptHeader
        )
        # EXCEPT    one
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.EXCEPT),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'one'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ExceptHeader,
            patterns=['one']
        )
        # EXCEPT    one    two    AS    ${var}
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.EXCEPT),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'one'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'two'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.AS, 'AS'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.VARIABLE, '${var}'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ExceptHeader,
            patterns=['one', 'two'],
            variable='${var}'
        )
        # EXCEPT    Example: *    type=glob
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.EXCEPT),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'Example: *'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.OPTION, 'type=glob'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            ExceptHeader,
            patterns=['Example: *'],
            type='glob'
        )
        # EXCEPT    Error \\d    (x|y)    type=regexp    AS    ${var}
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.EXCEPT),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'Error \\d'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '(x|y)'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.OPTION, 'type=regexp'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.AS, 'AS'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.VARIABLE, '${var}'),
            Token(Token.EOL, '\n')]
        assert_created_statement(
            tokens,
            ExceptHeader,
            patterns=['Error \\d', '(x|y)'],
            type='regexp',
            variable='${var}'
        )

    def test_FinallyHeader(self):
        # FINALLY
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.FINALLY),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            FinallyHeader
        )

    def test_WhileHeader(self):
        # WHILE    $cond
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.WHILE),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '$cond'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            WhileHeader,
            condition='$cond'
        )
        # WHILE    $cond    limit=100s
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.WHILE),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '$cond'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.OPTION, 'limit=100s'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            WhileHeader,
            condition='$cond',
            limit='100s'
        )
        # WHILE    $cond    limit=10    on_limit_message=Error message
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.WHILE),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, '$cond'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.OPTION, 'limit=10'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.OPTION, 'on_limit_message=Error message'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            WhileHeader,
            condition='$cond',
            limit='10',
            on_limit_message='Error message'
        )

    def test_End(self):
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.END),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            End
        )

    def test_ReturnStatement(self):
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.RETURN_STATEMENT),
            Token(Token.EOL)
        ]
        assert_created_statement(tokens, ReturnStatement)
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.RETURN_STATEMENT, 'RETURN'),
            Token(Token.SEPARATOR, '    '),
            Token(Token.ARGUMENT, 'x'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(tokens, ReturnStatement, values=('x',))

    def test_Break(self):
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.BREAK),
            Token(Token.EOL)
        ]
        assert_created_statement(tokens, Break)

    def test_Continue(self):
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.CONTINUE),
            Token(Token.EOL)
        ]
        assert_created_statement(tokens, Continue)

    def test_Comment(self):
        tokens = [
            Token(Token.SEPARATOR, '    '),
            Token(Token.COMMENT, '# example comment'),
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            Comment,
            comment='# example comment'
        )

    def test_EmptyLine(self):
        tokens = [
            Token(Token.EOL, '\n')
        ]
        assert_created_statement(
            tokens,
            EmptyLine,
            eol='\n'
        )


if __name__ == '__main__':
    unittest.main()



from typing import (Any, Dict, List, Mapping, MutableMapping, MutableSet, MutableSequence,
                    Set, Sequence, Tuple, Union)
try:
    from typing_extensions import TypedDict
except ImportError:
    from typing import TypedDict

from robot.api.deco import not_keyword


TypedDict = not_keyword(TypedDict)


class Point2D(TypedDict):
    x: int
    y: int


class Point(Point2D, total=False):
    z: int


class BadIntMeta(type(int)):
    def __instancecheck__(self, instance):
        raise TypeError('Bang!')


class BadInt(int, metaclass=BadIntMeta):
    pass


def list_(argument: List, expected=None):
    _validate_type(argument, expected)


def list_with_types(argument: List[int], expected=None, same=False):
    _validate_type(argument, expected, same)


def tuple_(argument: Tuple, expected=None):
    _validate_type(argument, expected)


def tuple_with_types(argument: Tuple[bool, int], expected=None, same=False):
    _validate_type(argument, expected, same)


def homogenous_tuple(argument: Tuple[int, ...], expected=None, same=False):
    _validate_type(argument, expected, same)


def sequence(argument: Sequence, expected=None):
    _validate_type(argument, expected)


def sequence_with_types(argument: Sequence[Union[int, float]], expected=None):
    _validate_type(argument, expected)


def mutable_sequence(argument: MutableSequence, expected=None):
    _validate_type(argument, expected)


def mutable_sequence_with_types(argument: MutableSequence[int], expected=None):
    _validate_type(argument, expected)


def dict_(argument: Dict, expected=None):
    _validate_type(argument, expected)


def dict_with_types(argument: Dict[int, float], expected=None, same=False):
    _validate_type(argument, expected, same)


def mapping(argument: Mapping, expected=None):
    _validate_type(argument, expected)


def mapping_with_types(argument: Mapping[int, float], expected=None):
    _validate_type(argument, expected)


def mutable_mapping(argument: MutableMapping, expected=None):
    _validate_type(argument, expected)


def mutable_mapping_with_types(argument: MutableMapping[int, float], expected=None):
    _validate_type(argument, expected)


def typeddict(argument: Point2D, expected=None):
    _validate_type(argument, expected)


def typeddict_with_optional(argument: Point, expected=None):
    _validate_type(argument, expected)


def set_(argument: Set, expected=None):
    _validate_type(argument, expected)


def set_with_types(argument: Set[int], expected=None, same=False):
    _validate_type(argument, expected, same)


def mutable_set(argument: MutableSet, expected=None):
    _validate_type(argument, expected)


def mutable_set_with_types(argument: MutableSet[float], expected=None):
    _validate_type(argument, expected)


def any_(argument: Any = 1, expected=None):
    _validate_type(argument, expected)


def none_as_default(argument: List = None, expected=None):
    _validate_type(argument, expected)


def none_as_default_with_any(argument: Any = None, expected=None):
    _validate_type(argument, expected)


def forward_reference(argument: 'List', expected=None):
    _validate_type(argument, expected)


def forward_ref_with_types(argument: 'List[int]', expected=None):
    _validate_type(argument, expected)


def not_liking_isinstance(argument: BadInt, expected=None):
    _validate_type(argument, expected)


def _validate_type(argument, expected, same=False):
    if isinstance(expected, str):
        expected = eval(expected)
    if argument != expected or type(argument) != type(expected):
        atype = type(argument).__name__
        etype = type(expected).__name__
        raise AssertionError(f'{argument!r} ({atype}) != {expected!r} ({etype})')
    if same and argument is not expected:
        raise AssertionError(f'{argument} (id: {id(argument)}) is not same '
                             f'as {expected} (id: {id(expected)})')



import inspect
import os.path
import re
import sys
import unittest

from robot.running.handlers import _PythonHandler, DynamicHandler
from robot.utils.asserts import assert_equal, assert_raises_with_msg, assert_true
from robot.running.testlibraries import TestLibrary, LibraryScope
from robot.running.dynamicmethods import (
    GetKeywordArguments, GetKeywordDocumentation, RunKeyword)
from robot.errors import DataError

from classes import (NameLibrary, DocLibrary, ArgInfoLibrary,
                     __file__ as classes_source)
from ArgumentsPython import ArgumentsPython


def _get_handler_methods(lib):
    attrs = [getattr(lib, a) for a in dir(lib) if not a.startswith('_')]
    return [a for a in attrs if inspect.ismethod(a)]


class LibraryMock:

    def __init__(self, name='MyLibrary', scope='GLOBAL'):
        self.name = self.orig_name = name
        self.scope = LibraryScope(scope, self)

    register_listeners = unregister_listeners = reset_instance \
        = get_instance = lambda *args: None


def assert_argspec(argspec, minargs=0, maxargs=0, positional=[], defaults={},
                   varargs=None, kwonlyargs=[], kwargs=None):
    assert_equal(argspec.minargs, minargs)
    assert_equal(argspec.maxargs, maxargs)
    assert_equal(argspec.positional, positional)
    assert_equal(argspec.defaults, defaults)
    assert_equal(argspec.var_positional, varargs)
    assert_equal(argspec.named_only, kwonlyargs)
    assert_equal(argspec.var_named, kwargs)


class TestPythonHandler(unittest.TestCase):

    def test_name(self):
        for method in _get_handler_methods(NameLibrary()):
            handler = _PythonHandler(LibraryMock('mylib'), method.__name__, method)
            assert_equal(handler.name, method.__doc__)
            assert_equal(handler.longname, 'mylib.'+method.__doc__)

    def test_docs(self):
        for method in _get_handler_methods(DocLibrary()):
            handler = _PythonHandler(LibraryMock(), method.__name__, method)
            assert_equal(handler.doc, method.expected_doc)
            assert_equal(handler.shortdoc, method.expected_shortdoc)

    def test_arguments(self):
        for method in _get_handler_methods(ArgInfoLibrary()):
            handler = _PythonHandler(LibraryMock(), method.__name__, method)
            args = handler.arguments
            argspec = (args.positional, args.defaults, args.var_positional, args.var_named)
            expected = eval(method.__doc__)
            assert_equal(argspec, expected, method.__name__)

    def test_arg_limits(self):
        for method in _get_handler_methods(ArgumentsPython()):
            handler = _PythonHandler(LibraryMock(), method.__name__, method)
            exp_mina, exp_maxa = eval(method.__doc__)
            assert_equal(handler.arguments.minargs, exp_mina)
            assert_equal(handler.arguments.maxargs, exp_maxa)

    def test_getarginfo_getattr(self):
        handlers = TestLibrary('classes.GetattrLibrary').handlers
        assert_equal(len(handlers), 3)
        for handler in handlers:
            assert_true(handler.name in ['Foo','Bar','Zap'])
            assert_equal(handler.arguments.minargs, 0)
            assert_equal(handler.arguments.maxargs, sys.maxsize)


class TestDynamicHandlerCreation(unittest.TestCase):

    def test_none_doc(self):
        self._assert_doc(None, '')

    def test_empty_doc(self):
        self._assert_doc('')

    def test_non_empty_doc(self):
        self._assert_doc('This is some documentation')

    def test_non_ascii_doc(self):
        self._assert_doc('Hyvää yötä')

    def test_with_utf8_doc(self):
        doc = 'Hyvää yötä'
        self._assert_doc(doc.encode('UTF-8'), doc)

    def test_invalid_doc_type(self):
        self._assert_fails('Return value must be a string, got boolean.', doc=True)

    def test_none_argspec(self):
        self._assert_spec(None, maxargs=sys.maxsize, varargs='varargs', kwargs=False)

    def test_none_argspec_when_kwargs_supported(self):
        self._assert_spec(None, maxargs=sys.maxsize, varargs='varargs', kwargs='kwargs')

    def test_empty_argspec(self):
        self._assert_spec([])

    def test_mandatory_args(self):
        for argspec in [['arg'], ['arg1', 'arg2', 'arg3']]:
            self._assert_spec(argspec, len(argspec), len(argspec), argspec)

    def test_only_default_args(self):
        self._assert_spec(['d1=default', 'd2=True'], 0, 2,
                          ['d1', 'd2'], {'d1': 'default', 'd2': 'True'})

    def test_default_as_tuple_or_list_like(self):
        self._assert_spec([('d1', 'default'), ['d2', True]], 0, 2,
                          ['d1', 'd2'], {'d1': 'default', 'd2': True})

    def test_default_value_may_contain_equal_sign(self):
        self._assert_spec(['d=foo=bar'], 0, 1, ['d'], {'d': 'foo=bar'})

    def test_default_value_as_tuple_may_contain_equal_sign(self):
        self._assert_spec([('n=m', 'd=f')], 0, 1, ['n=m'], {'n=m': 'd=f'})

    def test_varargs(self):
        self._assert_spec(['*vararg'], 0, sys.maxsize, varargs='vararg')

    def test_kwargs(self):
        self._assert_spec(['**kwarg'], 0, 0, kwargs='kwarg')

    def test_varargs_and_kwargs(self):
        self._assert_spec(['*vararg', '**kwarg'],
                          0, sys.maxsize, varargs='vararg', kwargs='kwarg')

    def test_kwonlyargs(self):
        self._assert_spec(['*', 'kwo'], kwonlyargs=['kwo'])
        self._assert_spec(['*vars', 'kwo'], varargs='vars', kwonlyargs=['kwo'])
        self._assert_spec(['*', 'x', 'y', 'z'], kwonlyargs=['x', 'y', 'z'])

    def test_kwonlydefaults(self):
        self._assert_spec(['*', 'kwo=default'],
                          kwonlyargs=['kwo'],
                          defaults={'kwo': 'default'})
        self._assert_spec(['*vars', 'kwo=default'],
                          varargs='vars',
                          kwonlyargs=['kwo'],
                          defaults={'kwo': 'default'})
        self._assert_spec(['*', 'x=1', 'y', 'z=3'],
                          kwonlyargs=['x', 'y', 'z'],
                          defaults={'x': '1', 'z': '3'})

    def test_kwonlydefaults_with_tuple(self):
        self._assert_spec(['*', ('kwo', 'default')],
                          kwonlyargs=['kwo'],
                          defaults={'kwo': 'default'})
        self._assert_spec([('*',), 'x=1', 'y', ('z', 3)],
                          kwonlyargs=['x', 'y', 'z'],
                          defaults={'x': '1', 'z': 3})

    def test_integration(self):
        self._assert_spec(['arg', 'default=value'],
                          1, 2,
                          positional=['arg', 'default'],
                          defaults={'default': 'value'})
        self._assert_spec(['arg', 'default=value', '*var'],
                          1, sys.maxsize,
                          positional=['arg', 'default'],
                          defaults={'default': 'value'},
                          varargs='var')
        self._assert_spec(['arg', 'default=value', '**kw'],
                          1, 2,
                          positional=['arg', 'default'],
                          defaults={'default': 'value'},
                          kwargs='kw')
        self._assert_spec(['arg', 'default=value', '*var', '**kw'],
                          1, sys.maxsize,
                          positional=['arg', 'default'],
                          defaults={'default': 'value'},
                          varargs='var',
                          kwargs='kw')
        self._assert_spec(['a', 'b=1', 'c=2', '*d', 'e', 'f=3', 'g', '**h'],
                          1, sys.maxsize,
                          positional=['a', 'b', 'c'],
                          defaults={'b': '1', 'c': '2', 'f': '3'},
                          varargs='d',
                          kwonlyargs=['e', 'f', 'g'],
                          kwargs='h')
        self._assert_spec([('a',), ('b', '1'), ('c', 2), ('*d',), ('e',), ('f', 3), ('g',), ('**h',)],
                          1, sys.maxsize,
                          positional=['a', 'b', 'c'],
                          defaults={'b': '1', 'c': 2, 'f': 3},
                          varargs='d',
                          kwonlyargs=['e', 'f', 'g'],
                          kwargs='h')

    def test_invalid_argspec_type(self):
        for argspec in [True, [1, 2], ['arg', ()]]:
            self._assert_fails("Return value must be a list of strings "
                               "or non-empty tuples.", argspec)

    def test_invalid_tuple(self):
        for invalid in [('too', 'many', 'values'), ('*too', 'many'),
                        ('**too', 'many'), (1, 2), (1,)]:
            self._assert_fails('Invalid argument specification: '
                               'Invalid argument "%s".' % (invalid,),
                               ['valid', invalid])

    def test_mandatory_arg_after_default_arg(self):
        for argspec in [['d=v', 'arg'], ['a', 'b', 'c=v', 'd']]:
            self._assert_fails('Invalid argument specification: '
                               'Non-default argument after default arguments.',
                               argspec)

    def test_multiple_vararg(self):
        self._assert_fails('Invalid argument specification: '
                           'Cannot have multiple varargs.',
                           ['*first', '*second'])

    def test_vararg_with_kwonly_separator(self):
        self._assert_fails('Invalid argument specification: '
                           'Cannot have multiple varargs.',
                           ['*', '*varargs'])
        self._assert_fails('Invalid argument specification: '
                           'Cannot have multiple varargs.',
                           ['*varargs', '*'])
        self._assert_fails('Invalid argument specification: '
                           'Cannot have multiple varargs.',
                           ['*', '*'])

    def test_kwarg_not_last(self):
        for argspec in [['**foo', 'arg'], ['arg', '**kw', 'arg'],
                        ['a', 'b=d', '**kw', 'c'], ['**kw', '*vararg'],
                        ['**kw', '**kwarg']]:
            self._assert_fails('Invalid argument specification: '
                               'Only last argument can be kwargs.', argspec)

    def test_missing_kwargs_support(self):
        for spec in (['**kwargs'], ['arg', '**kws'], ['a', '*v', '**k']):
            self._assert_fails("Too few 'run_keyword' method parameters "
                               "for **kwargs support.", spec)

    def test_missing_kwonlyargs_support(self):
        for spec in (['*', 'kwo'], ['*vars', 'kwo1', 'kwo2=default']):
            self._assert_fails("Too few 'run_keyword' method parameters "
                               "for keyword-only arguments support.", spec)

    def _assert_doc(self, doc, expected=None):
        expected = doc if expected is None else expected
        assert_equal(self._create_handler(doc=doc).doc, expected)

    def _assert_spec(self, argspec, minargs=0, maxargs=0,
                     positional=[], defaults={}, varargs=None,
                     kwonlyargs=[], kwargs=None):
        if varargs and not maxargs:
            maxargs = sys.maxsize
        if kwargs is None and not kwonlyargs:
            kwargs_support_modes = [True, False]
        elif kwargs is False:
            kwargs_support_modes = [False]
            kwargs = None
        else:
            kwargs_support_modes = [True]
        for kwargs_support in kwargs_support_modes:
            handler = self._create_handler(argspec, kwargs_support=kwargs_support)
            assert_argspec(handler.arguments, minargs, maxargs, positional,
                           defaults, varargs, kwonlyargs, kwargs)

    def _assert_fails(self, error, *args, **kwargs):
        assert_raises_with_msg(DataError, error,
                               self._create_handler, *args, **kwargs)

    def _create_handler(self, argspec=None, doc=None, kwargs_support=False):
        lib = LibraryMock('TEST CASE')
        if kwargs_support:
            lib.run_keyword = lambda name, args, kwargs: None
        else:
            lib.run_keyword = lambda name, args: None
        lib.run_keyword.__name__ = 'run_keyword'
        doc = GetKeywordDocumentation(lib)._handle_return_value(doc)
        argspec = GetKeywordArguments(lib)._handle_return_value(argspec)
        return DynamicHandler(lib, 'mock', RunKeyword(lib), doc, argspec)


class TestSourceAndLineno(unittest.TestCase):

    def test_class_with_init(self):
        lib = TestLibrary('classes.RecordingLibrary')
        self._verify(lib.handlers['kw'], classes_source, 206)
        self._verify(lib.init, classes_source, 202)

    def test_class_without_init(self):
        lib = TestLibrary('classes.NameLibrary')
        self._verify(lib.handlers['simple1'], classes_source, 13)
        self._verify(lib.init, classes_source, -1)

    def test_module(self):
        from module_library import __file__ as source
        lib = TestLibrary('module_library')
        self._verify(lib.handlers['passing'], source, 5)
        self._verify(lib.init, source, -1)

    def test_package(self):
        from robot.variables.search import __file__ as source
        from robot.variables import __file__ as init_source
        lib = TestLibrary('robot.variables')
        self._verify(lib.handlers['is_variable'], source, 33)
        self._verify(lib.init, init_source, -1)

    def test_decorated(self):
        lib = TestLibrary('classes.Decorated')
        self._verify(lib.handlers['no_wrapper'], classes_source, 320)
        self._verify(lib.handlers['wrapper'], classes_source, 327)
        self._verify(lib.handlers['external'], classes_source, 332)
        self._verify(lib.handlers['no_def'], classes_source, 335)

    def test_dynamic_without_source(self):
        lib = TestLibrary('classes.ArgDocDynamicLibrary')
        self._verify(lib.handlers['No Arg'], classes_source, -1)

    def test_dynamic(self):
        lib = TestLibrary('classes.DynamicWithSource')
        self._verify(lib.handlers['only path'],
                     classes_source)
        self._verify(lib.handlers['path & lineno'],
                     classes_source, 42)
        self._verify(lib.handlers['lineno only'],
                     classes_source, 6475)
        self._verify(lib.handlers['invalid path'],
                     'path validity is not validated')
        self._verify(lib.handlers['path w/ colon'],
                     r'c:\temp\lib.py', -1)
        self._verify(lib.handlers['path w/ colon & lineno'],
                     r'c:\temp\lib.py', 1234567890)
        self._verify(lib.handlers['no source'],
                     classes_source)

    def test_dynamic_with_non_ascii_source(self):
        lib = TestLibrary('classes.DynamicWithSource')
        self._verify(lib.handlers['nön-äscii'], 'hyvä esimerkki')
        self._verify(lib.handlers['nön-äscii utf-8'], '福', 88)

    def test_dynamic_init(self):
        lib_with_init = TestLibrary('classes.ArgDocDynamicLibrary')
        lib_without_init = TestLibrary('classes.DynamicWithSource')
        self._verify(lib_with_init.init, classes_source, 217)
        self._verify(lib_without_init.init, classes_source, -1)

    def test_dynamic_invalid_source(self):
        logger = LoggerMock()
        lib = TestLibrary('classes.DynamicWithSource', logger=logger)
        self._verify(lib.handlers['invalid source'], lib.source)
        error = (
            "Error in library 'classes.DynamicWithSource': "
            "Getting source information for keyword 'Invalid Source' failed: "
            "Calling dynamic method 'get_keyword_source' failed: "
            "Return value must be a string, got integer."
        )
        assert_equal(logger.messages, [(error, 'ERROR')])

    def _verify(self, kw, source, lineno=-1):
        if source:
            source = re.sub(r'(\.pyc|\$py\.class)$', '.py', source)
            source = os.path.normpath(source)
        assert_equal(kw.source, source)
        assert_equal(kw.lineno, lineno)


class LoggerMock:

    def __init__(self):
        self.messages = []

    def write(self, message, level):
        self.messages.append((message, level))

    def info(self, message):
        self.write(message, 'INFO')

    def debug(self, message):
        pass


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import os.path

from robot.utils import abspath, normpath, get_link_path, WINDOWS
from robot.utils.robotpath import CASE_INSENSITIVE_FILESYSTEM
from robot.utils.asserts import assert_equal, assert_true


class TestAbspathNormpath(unittest.TestCase):

    def test_abspath(self):
        for inp, exp in self._get_inputs():
            exp = os.path.abspath(exp)
            path = abspath(inp)
            assert_equal(path, exp, inp)
            assert_true(isinstance(path, str), inp)
            exp = exp.lower() if CASE_INSENSITIVE_FILESYSTEM else exp
            path = abspath(inp, case_normalize=True)
            assert_equal(path, exp, inp)
            assert_true(isinstance(path, str), inp)

    def test_abspath_when_cwd_is_non_ascii(self):
        orig = abspath('.')
        nonasc = u'\xe4'
        os.mkdir(nonasc)
        os.chdir(nonasc)
        try:
            assert_equal(abspath('.'), orig + os.sep + nonasc)
        finally:
            os.chdir('..')
            os.rmdir(nonasc)

    if WINDOWS:
        unc_path = r'\\server\D$\dir\.\f1\..\\f2'
        unc_exp = r'\\server\D$\dir\f2'

        def test_unc_path(self):
            assert_equal(abspath(self.unc_path), self.unc_exp)

        def test_unc_path_when_chdir_is_root(self):
            orig = abspath('.')
            os.chdir('\\')
            try:
                assert_equal(abspath(self.unc_path), self.unc_exp)
            finally:
                os.chdir(orig)

        def test_add_drive(self):
            drive = os.path.abspath(__file__)[:2]
            for path in ['.', os.path.basename(__file__), r'\abs\path']:
                assert_true(abspath(path).startswith(drive))

    def test_normpath(self):
        for inp, exp in self._get_inputs():
            path = normpath(inp)
            assert_equal(path, exp, inp)
            assert_true(isinstance(path, str), inp)
            exp = exp.lower() if CASE_INSENSITIVE_FILESYSTEM else exp
            path = normpath(inp, case_normalize=True)
            assert_equal(path, exp, inp)
            assert_true(isinstance(path, str), inp)

    def _get_inputs(self):
        inputs = self._windows_inputs if WINDOWS else self._posix_inputs
        for inp, exp in inputs():
            yield inp, exp
            if inp not in ['', os.sep]:
                for ext in [os.sep, os.sep+'.', os.sep+'.'+os.sep]:
                    yield inp + ext, exp
            if inp.endswith(os.sep):
                for ext in ['.', '.'+os.sep, '.'+os.sep+'.']:
                    yield inp + ext, exp
                yield inp + 'foo' + os.sep + '..', exp

    def _posix_inputs(self):
        return [('/tmp/', '/tmp'),
                ('/var/../opt/../tmp/.', '/tmp'),
                ('/non/Existing/..', '/non'),
                ('/', '/')] + self._generic_inputs()

    def _windows_inputs(self):
        inputs = [('c:\\temp', 'c:\\temp'),
                  ('C:\\TEMP\\', 'C:\\TEMP'),
                  ('C:\\xxx\\..\\yyy\\..\\temp\\.', 'C:\\temp'),
                  ('c:\\Non\\Existing\\..', 'c:\\Non')]
        for x in 'ABCDEFGHIJKLMNOPQRSTUVXYZ':
            base = '%s:\\' % x
            inputs.append((base, base))
            inputs.append((base.lower(), base.lower()))
            inputs.append((base[:2], base))
            inputs.append((base[:2].lower(), base.lower()))
            inputs.append((base+'\\foo\\..\\.\\BAR\\\\', base+'BAR'))
        inputs += [(inp.replace('/', '\\'), exp) for inp, exp in inputs]
        for inp, exp in self._generic_inputs():
            exp = exp.replace('/', '\\')
            inputs.extend([(inp, exp), (inp.replace('/', '\\'), exp)])
        return inputs

    def _generic_inputs(self):
        return [('', '.'),
                ('.', '.'),
                ('./', '.'),
                ('..', '..'),
                ('../', '..'),
                ('../..', '../..'),
                ('foo', 'foo'),
                ('foo/bar', 'foo/bar'),
                (u'\xe4', u'\xe4'),
                (u'\xe4/\xf6', u'\xe4/\xf6'),
                ('./foo', 'foo'),
                ('foo/.', 'foo'),
                ('foo/..', '.'),
                ('foo/../bar', 'bar'),
                ('foo/bar/zap/..', 'foo/bar')]


class TestGetLinkPath(unittest.TestCase):

    def test_basics(self):
        for base, target, expected in self._get_basic_inputs():
            assert_equal(get_link_path(target, base).replace('R:', 'r:'),
                         expected, '%s -> %s' % (target, base))

    def test_base_is_existing_file(self):
        assert_equal(get_link_path(os.path.dirname(__file__), __file__), '.')
        assert_equal(get_link_path(__file__, __file__),
                     self._expected_basename(__file__))

    def test_non_existing_paths(self):
        assert_equal(get_link_path('/nonex/target', '/nonex/base'), '../target')
        assert_equal(get_link_path('/nonex/t.ext', '/nonex/b.ext'), '../t.ext')
        assert_equal(get_link_path('/nonex', __file__),
                     os.path.relpath('/nonex', os.path.dirname(__file__)).replace(os.sep, '/'))

    def test_non_ascii_paths(self):
        assert_equal(get_link_path(u'\xe4\xf6.txt', ''), '%C3%A4%C3%B6.txt')
        assert_equal(get_link_path(u'\xe4/\xf6.txt', u'\xe4'), '%C3%B6.txt')

    def _get_basic_inputs(self):
        directory = os.path.dirname(__file__)
        inputs = [(directory, __file__, self._expected_basename(__file__)),
                  (directory, directory, '.'),
                  (directory, directory + '/', '.'),
                  (directory, directory + '//', '.'),
                  (directory, directory + '///', '.'),
                  (directory, directory + '/trailing/part', 'trailing/part'),
                  (directory, directory + '//trailing//part', 'trailing/part'),
                  (directory, directory + '/..', '..'),
                  (directory, directory + '/../X', '../X'),
                  (directory, directory + '/./.././/..', '../..'),
                  (directory, '.', os.path.relpath('.', directory).replace(os.sep, '/'))]
        platform_inputs = (self._posix_inputs() if os.sep == '/' else
                           self._windows_inputs())
        return inputs + platform_inputs

    def _expected_basename(self, path):
        return os.path.basename(path).replace('$py.class', '%24py.class')

    def _posix_inputs(self):
        return [('/tmp/', '/tmp/bar.txt', 'bar.txt'),
                ('/tmp', '/tmp/x/bar.txt', 'x/bar.txt'),
                ('/tmp/', '/tmp/x/y/bar.txt', 'x/y/bar.txt'),
                ('/tmp/', '/tmp/x/y/z/bar.txt', 'x/y/z/bar.txt'),
                ('/tmp', '/x/y/z/bar.txt', '../x/y/z/bar.txt'),
                ('/tmp/', '/x/y/z/bar.txt', '../x/y/z/bar.txt'),
                ('/tmp', '/x/bar.txt', '../x/bar.txt'),
                ('/tmp', '/x/y/z/bar.txt', '../x/y/z/bar.txt'),
                ('/', '/x/bar.txt', 'x/bar.txt'),
                ('/home//test', '/home/user', '../user'),
                ('//home/test', '/home/user', '../user'),
                ('///home/test', '/home/user', '../user'),
                ('////////////////home/test', '/home/user', '../user'),
                ('/path/to', '/path/to/result_in_same_dir.html',
                 'result_in_same_dir.html'),
                ('/path/to/dir', '/path/to/result_in_parent_dir.html',
                 '../result_in_parent_dir.html'),
                ('/path/to', '/path/to/dir/result_in_sub_dir.html',
                 'dir/result_in_sub_dir.html'),
                ('/commonprefix/sucks/baR', '/commonprefix/sucks/baZ.txt',
                 '../baZ.txt'),
                ('/a/very/long/path', '/no/depth/limitation',
                 '../../../../no/depth/limitation'),
                ('/etc/hosts', '/path/to/existing/file',
                 '../path/to/existing/file'),
                ('/path/to/identity', '/path/to/identity', '.')]

    def _windows_inputs(self):
        return [('c:\\temp\\', 'c:\\temp\\bar.txt', 'bar.txt'),
                ('c:\\temp', 'c:\\temp\\x\\bar.txt', 'x/bar.txt'),
                ('c:\\temp\\', 'c:\\temp\\x\\y\\bar.txt', 'x/y/bar.txt'),
                ('c:\\temp', 'c:\\temp\\x\\y\\z\\bar.txt', 'x/y/z/bar.txt'),
                ('c:\\temp\\', 'c:\\x\\y\\bar.txt', '../x/y/bar.txt'),
                ('c:\\temp', 'c:\\x\\y\\bar.txt', '../x/y/bar.txt'),
                ('c:\\temp', 'c:\\x\\bar.txt', '../x/bar.txt'),
                ('c:\\temp', 'c:\\x\\y\\z\\bar.txt', '../x/y/z/bar.txt'),
                ('c:\\temp\\', 'r:\\x\\y\\bar.txt', 'file:///r:/x/y/bar.txt'),
                ('c:\\', 'c:\\x\\bar.txt', 'x/bar.txt'),
                ('c:\\path\\to', 'c:\\path\\to\\result_in_same_dir.html',
                 'result_in_same_dir.html'),
                ('c:\\path\\to\\dir', 'c:\\path\\to\\result_in_parent.dir',
                 '../result_in_parent.dir'),
                ('c:\\path\\to', 'c:\\path\\to\\dir\\result_in_sub_dir.html',
                 'dir/result_in_sub_dir.html'),
                ('c:\\commonprefix\\sucks\\baR',
                 'c:\\commonprefix\\sucks\\baZ.txt', '../baZ.txt'),
                ('c:\\a\\very\\long\\path', 'c:\\no\\depth\\limitation',
                 '../../../../no/depth/limitation'),
                ('c:\\windows\\explorer.exe',
                 'c:\\windows\\path\\to\\existing\\file',
                 'path/to/existing/file'),
                ('c:\\path\\2\\identity', 'c:\\path\\2\\identity', '.')]


if __name__ == '__main__':
    unittest.main()



import copy
import json
import os
import tempfile
import unittest
import warnings
from pathlib import Path

from jsonschema import Draft202012Validator

from robot import api, model
from robot.model.modelobject import ModelObject
from robot.parsing import get_resource_model
from robot.running import (Break, Continue, Error, For, If, IfBranch, Keyword,
                           Return, ResourceFile, TestCase, TestDefaults, TestSuite,
                           Try, TryBranch, While)
from robot.running.model import UserKeyword
from robot.utils.asserts import (assert_equal, assert_false, assert_not_equal,
                                 assert_raises, assert_true)


CURDIR = Path(__file__).resolve().parent
MISCDIR = (CURDIR / '../../atest/testdata/misc').resolve()


class TestModelTypes(unittest.TestCase):

    def test_suite_setup_and_teardown(self):
        suite = TestSuite()
        assert_equal(type(suite.setup), Keyword)
        assert_equal(type(suite.teardown), Keyword)
        assert_not_equal(type(suite.setup), model.Keyword)
        assert_not_equal(type(suite.teardown), model.Keyword)

    def test_suite_test_case(self):
        test = TestSuite().tests.create()
        assert_equal(type(test), TestCase)
        assert_not_equal(type(test), model.TestCase)

    def test_test_case_keyword(self):
        kw = TestCase().body.create_keyword()
        assert_equal(type(kw), Keyword)
        assert_not_equal(type(kw), model.Keyword)


class TestUserKeyword(unittest.TestCase):

    def test_keywords_deprecation(self):
        uk = UserKeyword('Name')
        uk.body.create_keyword()
        uk.teardown.config(name='T')
        with warnings.catch_warnings(record=True) as w:
            kws = uk.keywords
            assert_equal(len(kws), 2)
            assert_true('deprecated' in str(w[0].message))
        assert_raises(AttributeError, kws.append, Keyword())
        assert_raises(AttributeError, setattr, uk, 'keywords', [])


class TestSuiteFromSources(unittest.TestCase):
    path = Path(os.getenv('TEMPDIR') or tempfile.gettempdir(),
                'test_run_model.robot')
    data = '''
*** Settings ***
Documentation    Some text.
Test Setup       No Operation
Library          ExampleLibrary

*** Variables ***
${VAR}           Value

*** Test Cases ***
Example
    [Tags]    tag
    Keyword

*** Keywords ***
Keyword
    Log    ${CURDIR}
'''

    @classmethod
    def setUpClass(cls):
        with open(cls.path, 'w') as f:
            f.write(cls.data)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.path)

    def test_from_file_system(self):
        suite = TestSuite.from_file_system(self.path)
        self._verify_suite(suite, curdir=str(self.path.parent))

    def test_from_file_system_with_multiple_paths(self):
        suite = TestSuite.from_file_system(self.path, self.path)
        assert_equal(suite.name, 'Test Run Model & Test Run Model')
        self._verify_suite(suite.suites[0], curdir=str(self.path.parent))
        self._verify_suite(suite.suites[1], curdir=str(self.path.parent))

    def test_from_file_system_with_config(self):
        suite = TestSuite.from_file_system(self.path, process_curdir=False)
        self._verify_suite(suite)

    def test_from_file_system_with_defaults(self):
        defaults = TestDefaults(tags=('from defaults',), timeout='10s')
        suite = TestSuite.from_file_system(self.path, defaults=defaults)
        self._verify_suite(suite, tags=('from defaults', 'tag'), timeout='10s',
                           curdir=str(self.path.parent))

    def test_from_model(self):
        model = api.get_model(self.data)
        suite = TestSuite.from_model(model)
        self._verify_suite(suite, name='')

    def test_from_model_containing_source(self):
        model = api.get_model(self.path)
        suite = TestSuite.from_model(model)
        self._verify_suite(suite)

    def test_from_model_with_defaults(self):
        model = api.get_model(self.path)
        defaults = TestDefaults(tags=('from defaults',), timeout='10s')
        suite = TestSuite.from_model(model, defaults=defaults)
        self._verify_suite(suite, tags=('from defaults', 'tag'), timeout='10s')

    def test_from_model_with_custom_name(self):
        for source in [self.data, self.path]:
            model = api.get_model(source)
            with warnings.catch_warnings(record=True) as w:
                suite = TestSuite.from_model(model, name='Custom name')
                assert_equal(str(w[0].message),
                             "'name' argument of 'TestSuite.from_model' is deprecated. "
                             "Set the name to the returned suite separately.")
            self._verify_suite(suite, 'Custom name')

    def test_from_string(self):
        suite = TestSuite.from_string(self.data)
        self._verify_suite(suite, name='')

    def test_from_string_with_config(self):
        suite = TestSuite.from_string(self.data.replace('Test Cases', 'Testit'),
                                      lang='Finnish', curdir='.')
        self._verify_suite(suite, name='', curdir='.')

    def test_from_string_with_defaults(self):
        defaults = TestDefaults(tags=('from defaults',), timeout='10s')
        suite = TestSuite.from_string(self.data, defaults=defaults)
        self._verify_suite(suite, name='', tags=('from defaults', 'tag'), timeout='10s')

    def _verify_suite(self, suite, name='Test Run Model', tags=('tag',),
                      timeout=None, curdir='${CURDIR}'):
        curdir = curdir.replace('\\', '\\\\')
        assert_equal(suite.name, name)
        assert_equal(suite.doc, 'Some text.')
        assert_equal(suite.rpa, False)
        assert_equal(suite.resource.imports[0].type, 'LIBRARY')
        assert_equal(suite.resource.imports[0].name, 'ExampleLibrary')
        assert_equal(suite.resource.variables[0].name, '${VAR}')
        assert_equal(suite.resource.variables[0].value, ('Value',))
        assert_equal(suite.resource.keywords[0].name, 'Keyword')
        assert_equal(suite.resource.keywords[0].body[0].name, 'Log')
        assert_equal(suite.resource.keywords[0].body[0].args, (curdir,))
        assert_equal(suite.tests[0].name, 'Example')
        assert_equal(suite.tests[0].tags, tags)
        assert_equal(suite.tests[0].timeout, timeout)
        assert_equal(suite.tests[0].setup.name, 'No Operation')
        assert_equal(suite.tests[0].body[0].name, 'Keyword')


class TestCopy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.suite = TestSuite.from_file_system(MISCDIR)

    def test_copy(self):
        self.assert_copy(self.suite, self.suite.copy())

    def assert_copy(self, original, copied):
        assert_not_equal(id(original), id(copied))
        self.assert_same_attrs_and_values(original, copied)
        for attr in ['suites', 'tests']:
            for child in getattr(original, attr, []):
                self.assert_copy(child, child.copy())

    def assert_same_attrs_and_values(self, model1, model2):
        assert_equal(dir(model1), dir(model2))
        for attr, value1, value2 in self.get_non_property_attrs(model1, model2):
            if callable(value1) and callable(value2):
                continue
            assert_equal(id(value1), id(value2), attr)
            if isinstance(value1, ModelObject):
                self.assert_same_attrs_and_values(value1, value2)

    def get_non_property_attrs(self, model1, model2):
        for attr in dir(model1):
            if 'parent' in attr or isinstance(getattr(type(model1), attr, None), property):
                continue
            value1 = getattr(model1, attr)
            value2 = getattr(model2, attr)
            yield attr, value1, value2

    def test_deepcopy(self):
        self.assert_deepcopy(self.suite, self.suite.deepcopy())

    def assert_deepcopy(self, original, copied):
        assert_not_equal(id(original), id(copied))
        self.assert_same_attrs_and_different_values(original, copied)
        # It would be too slow to test deepcopy recursively like we test copy.

    def assert_same_attrs_and_different_values(self, model1, model2):
        assert_equal(dir(model1), dir(model2))
        for attr, value1, value2 in self.get_non_property_attrs(model1, model2):
            if attr.startswith('__') or self.cannot_differ(value1, value2):
                continue
            assert_not_equal(id(value1), id(value2), attr)
            if isinstance(value1, ModelObject):
                self.assert_same_attrs_and_different_values(value1, value2)

    def cannot_differ(self, value1, value2):
        if isinstance(value1, ModelObject):
            return False
        if type(value1) is not type(value2):
            return False
        # None, Booleans, small numbers, etc. are singletons.
        try:
            return id(value1) == id(copy.deepcopy(value1))
        except TypeError:  # Got in some cases at least with Python 2.6
            return True


class TestLineNumberAndSource(unittest.TestCase):
    source = MISCDIR / 'pass_and_fail.robot'

    @classmethod
    def setUpClass(cls):
        cls.suite = TestSuite.from_file_system(cls.source)

    def test_suite(self):
        assert_equal(self.suite.source, self.source)
        assert_false(hasattr(self.suite, 'lineno'))

    def test_import(self):
        self._assert_lineno_and_source(self.suite.resource.imports[0], 5)

    def test_import_without_source(self):
        suite = TestSuite()
        suite.resource.imports.library('Example')
        assert_equal(suite.resource.imports[0].source, None)
        assert_equal(suite.resource.imports[0].directory, None)

    def test_import_with_non_existing_source(self):
        for source in Path('dummy!'), Path('dummy/example/path'):
            suite = TestSuite(source=source)
            suite.resource.imports.library('Example')
            assert_equal(suite.resource.imports[0].source, source)
            assert_equal(suite.resource.imports[0].directory, source.parent)

    def test_variable(self):
        self._assert_lineno_and_source(self.suite.resource.variables[0], 8)

    def test_test(self):
        self._assert_lineno_and_source(self.suite.tests[0], 12)

    def test_user_keyword(self):
        self._assert_lineno_and_source(self.suite.resource.keywords[0], 24)

    def test_keyword_call(self):
        self._assert_lineno_and_source(self.suite.tests[0].body[0], 15)
        self._assert_lineno_and_source(self.suite.resource.keywords[0].body[0], 27)

    def _assert_lineno_and_source(self, item, lineno):
        assert_equal(item.source, self.source)
        assert_equal(item.lineno, lineno)


class TestToFromDictAndJson(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open(CURDIR / '../../doc/schema/running.json') as file:
            schema = json.load(file)
        cls.validator = Draft202012Validator(schema=schema)

    def test_keyword(self):
        self._verify(Keyword(), name='')
        self._verify(Keyword('Name'), name='Name')
        self._verify(Keyword('N', 'args', ('${result}',)),
                     name='N', args=tuple('args'), assign=('${result}',))
        self._verify(Keyword('Setup', type=Keyword.SETUP, lineno=1),
                     name='Setup', lineno=1)

    def test_for(self):
        self._verify(For(), type='FOR', variables=(), flavor='IN', values=(), body=[])
        self._verify(For(['${i}'], 'IN RANGE', ['10'], lineno=2),
                     type='FOR', variables=('${i}',), flavor='IN RANGE', values=('10',),
                     body=[], lineno=2)
        self._verify(For(['${i}', '${a}'], 'IN ENUMERATE', ['cat', 'dog'], start='1'),
                     type='FOR', variables=('${i}', '${a}'), flavor='IN ENUMERATE',
                     values=('cat', 'dog'), start='1', body=[])

    def test_while(self):
        self._verify(While(), type='WHILE', body=[])
        self._verify(While('1 > 0', '1 min'),
                     type='WHILE', condition='1 > 0', limit='1 min', body=[])
        self._verify(While(limit='1', on_limit='PASS'),
                     type='WHILE', limit='1', on_limit='PASS', body=[])
        self._verify(While(limit='1', on_limit_message='Ooops!'),
                     type='WHILE', limit='1', on_limit_message='Ooops!', body=[])
        self._verify(While('True', lineno=3, error='x'),
                     type='WHILE', condition='True', body=[], lineno=3, error='x')

    def test_while_structure(self):
        root = While('True')
        root.body.create_keyword('K', 'a')
        root.body.create_while('False').body.create_keyword('W')
        root.body.create_break()
        self._verify(root, type='WHILE', condition='True',
                     body=[{'name': 'K', 'args': ('a',)},
                           {'type': 'WHILE', 'condition': 'False',
                            'body': [{'name': 'W'}]},
                           {'type': 'BREAK'}])

    def test_if(self):
        self._verify(If(), type='IF/ELSE ROOT', body=[])
        self._verify(If(lineno=4, error='E'),
                     type='IF/ELSE ROOT', body=[], lineno=4, error='E')

    def test_if_branch(self):
        self._verify(IfBranch(), type='IF', body=[])
        self._verify(IfBranch(If.ELSE_IF, '1 > 0'),
                     type='ELSE IF', condition='1 > 0', body=[])
        self._verify(IfBranch(If.ELSE, lineno=5),
                     type='ELSE', body=[], lineno=5)

    def test_if_structure(self):
        root = If()
        root.body.create_branch(If.IF, '$c').body.create_keyword('K1')
        root.body.create_branch(If.ELSE).body.create_keyword('K2', ['a'])
        self._verify(root,
                     type='IF/ELSE ROOT',
                     body=[{'type': 'IF', 'condition': '$c', 'body': [{'name': 'K1'}]},
                           {'type': 'ELSE', 'body': [{'name': 'K2', 'args': ('a',)}]}])

    def test_try(self):
        self._verify(Try(), type='TRY/EXCEPT ROOT', body=[])
        self._verify(Try(lineno=6, error='E'),
                     type='TRY/EXCEPT ROOT', body=[], lineno=6, error='E')

    def test_try_branch(self):
        self._verify(TryBranch(), type='TRY', body=[])
        self._verify(TryBranch(Try.EXCEPT), type='EXCEPT', patterns=(), body=[])
        self._verify(TryBranch(Try.EXCEPT, ['Pa*'], 'glob', '${err}'), type='EXCEPT',
                     patterns=('Pa*',), pattern_type='glob', variable='${err}', body=[])
        self._verify(TryBranch(Try.ELSE, lineno=7), type='ELSE', body=[], lineno=7)
        self._verify(TryBranch(Try.FINALLY, lineno=8), type='FINALLY', body=[], lineno=8)

    def test_try_structure(self):
        root = Try()
        root.body.create_branch(Try.TRY).body.create_keyword('K1')
        root.body.create_branch(Try.EXCEPT).body.create_keyword('K2')
        root.body.create_branch(Try.ELSE).body.create_keyword('K3')
        root.body.create_branch(Try.FINALLY).body.create_keyword('K4')
        self._verify(root,
                     type='TRY/EXCEPT ROOT',
                     body=[{'type': 'TRY', 'body': [{'name': 'K1'}]},
                           {'type': 'EXCEPT', 'patterns': (), 'body': [{'name': 'K2'}]},
                           {'type': 'ELSE', 'body': [{'name': 'K3'}]},
                           {'type': 'FINALLY', 'body': [{'name': 'K4'}]}])

    def test_return_continue_break(self):
        self._verify(Return(), type='RETURN', values=())
        self._verify(Return(('x', 'y'), lineno=9, error='E'),
                     type='RETURN', values=('x', 'y'), lineno=9, error='E')
        self._verify(Continue(), type='CONTINUE')
        self._verify(Continue(lineno=10, error='E'),
                     type='CONTINUE', lineno=10, error='E')
        self._verify(Break(), type='BREAK')
        self._verify(Break(lineno=11, error='E'),
                     type='BREAK', lineno=11, error='E')

    def test_error(self):
        self._verify(Error(), type='ERROR', values=(), error='')
        self._verify(Error(('bad', 'things'), error='Bad things!'),
                     type='ERROR', values=('bad', 'things'), error='Bad things!')

    def test_test(self):
        self._verify(TestCase(), name='', body=[])
        self._verify(TestCase('N', 'D', 'T', '1s', lineno=12),
                     name='N', doc='D', tags=('T',), timeout='1s', lineno=12, body=[])
        self._verify(TestCase(template='K'), name='', body=[], template='K')

    def test_test_structure(self):
        test = TestCase('TC')
        test.setup.config(name='Setup')
        test.teardown.config(name='Teardown', args='a')
        test.body.create_keyword('K1', 'a')
        test.body.create_if().body.create_branch('IF', '$c').body.create_keyword('K2')
        self._verify(test,
                     name='TC',
                     setup={'name': 'Setup'},
                     teardown={'name': 'Teardown', 'args': ('a',)},
                     body=[{'name': 'K1', 'args': ('a',)},
                           {'type': 'IF/ELSE ROOT',
                            'body': [{'type': 'IF', 'condition': '$c',
                                      'body': [{'name': 'K2'}]}]}])

    def test_suite(self):
        self._verify(TestSuite(), name='', resource={})
        self._verify(TestSuite('N', 'D', {'M': 'V'}, 'x.robot', rpa=True),
                     name='N', doc='D', metadata={'M': 'V'}, source='x.robot', rpa=True,
                     resource={})

    def test_suite_structure(self):
        suite = TestSuite('Root')
        suite.setup.config(name='Setup')
        suite.teardown.config(name='Teardown', args='a')
        suite.tests.create('T1').body.create_keyword('K')
        suite.suites.create('Child').tests.create('T2')
        self._verify(suite,
                     name='Root',
                     setup={'name': 'Setup'},
                     teardown={'name': 'Teardown', 'args': ('a',)},
                     tests=[{'name': 'T1', 'body': [{'name': 'K'}]}],
                     suites=[{'name': 'Child',
                              'tests': [{'name': 'T2', 'body': []}],
                              'resource': {}}],
                     resource={})

    def test_user_keyword(self):
        self._verify(UserKeyword(), name='', body=[])
        self._verify(UserKeyword('N', ('a',), 'd', 't', ('r',), 't', 1, error='E'),
                     name='N',
                     args=('a',),
                     doc='d',
                     tags=('t',),
                     return_=('r',),
                     timeout='t',
                     lineno=1,
                     error='E',
                     body=[])

    def test_user_keyword_structure(self):
        uk = UserKeyword('UK')
        uk.body.create_keyword('K1')
        uk.body.create_if().body.create_branch(condition='$c').body.create_keyword('K2')
        uk.teardown.config(name='Teardown')
        self._verify(uk, name='UK',
                     body=[{'name': 'K1'},
                           {'type': 'IF/ELSE ROOT',
                            'body': [{'type': 'IF', 'condition': '$c',
                                      'body': [{'name': 'K2'}]}]}],
                     teardown={'name': 'Teardown'})

    def test_resource_file(self):
        self._verify(ResourceFile())
        resource = ResourceFile('x.resource', doc='doc')
        resource.imports.library('L', ['a'], 'A', 1)
        resource.imports.resource('R', 2)
        resource.imports.variables('V', ['a'], 3)
        resource.variables.create('${x}', ('value',))
        resource.variables.create('@{y}', ('v1', 'v2'), lineno=4)
        resource.variables.create('&{z}', ['k=v'], error='E')
        resource.keywords.create('UK').body.create_keyword('K')
        self._verify(resource,
                     source='x.resource',
                     doc='doc',
                     imports=[{'type': 'LIBRARY', 'name': 'L', 'args': ('a',),
                               'alias': 'A', 'lineno': 1},
                              {'type': 'RESOURCE', 'name': 'R', 'lineno': 2},
                              {'type': 'VARIABLES', 'name': 'V', 'args': ('a',),
                               'lineno': 3}],
                     variables=[{'name': '${x}', 'value': ('value',)},
                                {'name': '@{y}', 'value': ('v1', 'v2'), 'lineno': 4},
                                {'name': '&{z}', 'value': ('k=v',), 'error': 'E'}],
                     keywords=[{'name': 'UK', 'body': [{'name': 'K'}]}])

    def test_bigger_suite_structure(self):
        suite = TestSuite.from_file_system(MISCDIR)
        self._verify(suite, **suite.to_dict())

    def _verify(self, obj, **expected):
        data = obj.to_dict()
        self.assertListEqual(list(data), list(expected))
        self.assertDictEqual(data, expected)
        roundtrip = type(obj).from_dict(data).to_dict()
        self.assertDictEqual(roundtrip, expected)
        roundtrip = type(obj).from_json(obj.to_json()).to_dict()
        self.assertDictEqual(roundtrip, expected)
        self._validate(obj)

    def _validate(self, obj):
        suite = self._create_suite_structure(obj)
        self.validator.validate(instance=json.loads(suite.to_json()))
        # Validating `suite.to_dict` directly doesn't work due to tuples not
        # being accepted as arrays:
        # https://github.com/python-jsonschema/jsonschema/issues/148
        #self.validator.validate(instance=suite.to_dict())

    def _create_suite_structure(self, obj):
        suite = TestSuite()
        test = suite.tests.create()
        if isinstance(obj, TestSuite):
            suite = obj
        elif isinstance(obj, TestCase):
            suite.tests = [obj]
        elif isinstance(obj, (Keyword, For, While, If, Try, Error)):
            test.body.append(obj)
        elif isinstance(obj, (IfBranch, TryBranch)):
            item = If() if isinstance(obj, IfBranch) else Try()
            item.body.append(obj)
            test.body.append(item)
        elif isinstance(obj, (Break, Continue, Return)):
            branch = test.body.create_if().body.create_branch()
            branch.body.append(obj)
        elif isinstance(obj, UserKeyword):
            suite.resource.keywords.append(obj)
        elif isinstance(obj, ResourceFile):
            suite.resource = obj
        else:
            raise ValueError(obj)
        return suite


class TestResourceFile(unittest.TestCase):
    path = CURDIR.parent / 'resources/test.resource'
    data = '''
*** Settings ***
Library         Example
Keyword Tags    common

*** Variables ***
${NAME}         Value

*** Keywords ***
Example
    [Tags]    own
    Log    Hello!
'''

    def test_from_file_system(self):
        res = ResourceFile.from_file_system(self.path)
        assert_equal(res.variables[0].name, '${PATH}')
        assert_equal(res.variables[0].value, (str(self.path.parent).replace('\\', '\\\\'),))
        assert_equal(res.keywords[0].name, 'My Test Keyword')

    def test_from_file_system_with_config(self):
        res = ResourceFile.from_file_system(self.path, process_curdir=False)
        assert_equal(res.variables[0].name, '${PATH}')
        assert_equal(res.variables[0].value, ('${CURDIR}',))
        assert_equal(res.keywords[0].name, 'My Test Keyword')

    def test_from_string(self):
        res = ResourceFile.from_string(self.data)
        assert_equal(res.imports[0].name, 'Example')
        assert_equal(res.variables[0].name, '${NAME}')
        assert_equal(res.variables[0].value, ('Value',))
        assert_equal(res.keywords[0].name, 'Example')
        assert_equal(res.keywords[0].tags, ['common', 'own'])
        assert_equal(res.keywords[0].body[0].name, 'Log')
        assert_equal(res.keywords[0].body[0].args, ('Hello!',))

    def test_from_string_with_config(self):
        res = ResourceFile.from_string('*** Muuttujat ***\n${NIMI}\tarvo', lang='fi')
        assert_equal(res.variables[0].name, '${NIMI}')
        assert_equal(res.variables[0].value, ('arvo',))

    def test_from_model(self):
        model = get_resource_model(self.data)
        res = ResourceFile.from_model(model)
        assert_equal(res.imports[0].name, 'Example')
        assert_equal(res.variables[0].name, '${NAME}')
        assert_equal(res.variables[0].value, ('Value',))
        assert_equal(res.keywords[0].name, 'Example')
        assert_equal(res.keywords[0].tags, ['common', 'own'])
        assert_equal(res.keywords[0].body[0].name, 'Log')
        assert_equal(res.keywords[0].body[0].args, ('Hello!',))


if __name__ == '__main__':
    unittest.main()
